/*  *********************************************************************** */
/*  *********************************************************************** */

/*  FILE set.h - *SAT 1.1. */
/*  Memoizing with sets of sets */

/*  *********************************************************************** */
/*  *********************************************************************** */

#include "util.h"
#include "var_set.h"

#include "dp_sat.h"

#ifndef MEMO
#define MEMO

/* Flags for the data structure 
*/
#define MAX_FLAGS 5        /* No more than 5 flags */
#define MAX_STATS 7        /* No more than 7 statistics */

/* Flags */
#define COMPACT_STORAGE  0 /* Try to factorize storage */
#define TRASH_SUBSETS    1 /* Try to trash subsets only */
#define GET_DEPENDENCIES 2 /* Store dependencies */

/* Statistics */
#define MEMO_ACCESS   0   /* Number of times the cache is searched */
#define MEMO_SUCCESS  1   /* Number of times the cache is useful */
#define DEP_ACCESS    2   /* Number of times a dependency is sought */
#define DEP_SUCCESS   3   /* Number of times a dependency is found */
#define MEMORY        4   /* Bytes allocated */
#define MEMO_STORAGE  5   /* Number of sets stored */
#define DEP_STORAGE   6   /* Number of dependencies stored */

/* The memoizing data structure
 */
typedef struct memo {
  var_set_t **memo_array;        /* the data sets */
  int       cur_set;             /* index to the next data set */
  int       growth;              /* 0 = fast growth, 1 = slow growth */  
  
  var_set_t **dep_array;         /* the dependency sets */
  int       *dep_beta;           /* the dependency labels */
  int       cur_dep;             /* index to the next dependency set */

  int       dim;                 /* initial dimensions */ 
  int       flags[MAX_FLAGS];    /* flags */
  int       stats[MAX_STATS];    /* statistics */     
} memo_t;

/* *********************************************************************** */
/* Initialize and clear memoizing */
/* *********************************************************************** */
void Init_memoS(int dim);
void Clear_memoS(int dim);
void Force_no_memoS();

/* *********************************************************************** */
/* Accessing the structure */
/* *********************************************************************** */
/* Test if the structure is being used; return 1 if so, 0 otherwise */
int Use_memoS();

/* Set & get flags */
void Setf_memoS(int flag);
void Clearf_memoS(int flag);
int  Getf_memoS(int flag);

/* Get stats */
int Stats_memoS(int stat);

/* Returns a sublist of beta_lsList_ptr containing betas that are not 
   tested with the current set of alphas yet. 
*/
lsList *Test_memoS(lsList *alpha_lsList_ptr, lsList *beta_lsList_ptr);

/* Stores a list of alphas that were tested with the 
   corresponding beta. 
   If COMPACT_STORAGE is set, the routine assumes that the beta should be 
   added in the current set. alphas ARE NOT CHECKED!!!
   If TRASH_SUBSETS is set, the routine alternates a "fast" (destructive)
   growth mode and a "slow" (conservative) growth mode. */
void Store_memoS(lsList *alpha_lsList_ptr, vertex_t *beta_vertex_ptr);

/* Returns a sublist of alpha_lsList_ptr containing dependency-free alphas */
lsList *Test_depS(lsList *alpha_lsList_ptr);

/* Remembers that the list of alphas implies the beta */
void Store_depS(lsList *alpha_lsList_ptr, vertex_t *beta_vertex_ptr); 

#endif






